"""
Django settings for coe project.

Generated by 'django-admin startproject' using Django 2.2.6.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""

SECRET_KEY = '2@ttqh(gormw%0gf6fb!%h381p=%+-!^*23qx#m@6nyg4%fh8v'

INSTALLED_APPS = [
    'django_extensions',
    'main',
    'seaflow'
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.mysql',
        'NAME': 'test',
        'USER': 'root',
        'PASSWORD': 'chenjian',
        'HOST': '127.0.0.1',
        'PORT': '3306',
        'OPTIONS': {'charset': 'utf8mb4'}
    },
}

# Celery
CELERY_ACCEPT_CONTENT = ['json']
CELERY_RESULT_SERIALIZER = 'json'
CELERY_TASK_SERIALIZER = 'json'
CELERY_BROKER_URL = 'redis://127.0.0.1:6379/6'
CELERY_RESULT_BACKEND = 'redis://127.0.0.1:6379/6'
